#  Author: Jonathan Michalon <michalon@igbmc.fr>, 2016
#  Institut de Génétique et de Biologie Moléculaire et Cellulaire
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#  
from PyQt5.QtCore import *
from PyQt5.QtWidgets import *


def alert(msg):
    msgBox = QMessageBox()
    msgBox.setText(msg)
    msgBox.exec()

def myfield(label, suffix, value, layout):
    from PyQt5.QtWidgets import QLabel, QWidget, QHBoxLayout
    from PyQt5.QtCore import Qt
    from gridimage import GridImage
    label = QLabel(label + ":")
    widget = QWidget()
    if type(value) in [QTextEdit, GridImage]:
        mylayout = QVBoxLayout()
    else:
        mylayout = QHBoxLayout()
    widget.setLayout(mylayout)
    mylayout.addWidget(label)
    mylayout.setAlignment(label, Qt.AlignTop)
    mylayout.addWidget(value)
    layout.addWidget(widget)
    
    if hasattr(value, "setAlignment") and type(value) != QTextEdit:
        value.setAlignment(Qt.AlignRight)
    if hasattr(value, "setSuffix"):
        value.setSuffix(" " + suffix)
    
    return value

def getimage(location, directory, name):
    import os
    
    if location.startswith('http://'):
        from dav import pull
        from io import BytesIO
        buf = BytesIO()
        pull(name, location, path=directory, buf=buf)
    else:
        buf = open(os.path.join(location, directory, name), 'rb')
    
    return buf

def addimage(fname, location, directory, name):
    import os
    
    data = open(os.path.normcase(fname), 'rb')
    if location.startswith('http://'):
        from dav import push
        push(name, location, path=directory, buf=data)
    else:
        import shutil
        if not os.access(os.path.join(location, directory), os.W_OK):
            os.makedirs(os.path.join(location, directory))
        out = open(os.path.join(location, directory, name), 'wb')
        shutil.copyfileobj(data, out, length=16*1024*1024)
    
    data.seek(0)
    return data

def moveimage(location, directory, image, newname):
    import os

    split = image['filename'].rsplit('-', 2)
    typeext = split[1] if len(split) > 1 else ''
    newname = newname + '-' + typeext

    if location.startswith('http://'):
        from dav import move
        move(image['filename'], location, newname, path=directory)
    else:
        os.rename(os.path.join(location, directory, image['filename']), os.path.join(location, directory, newname))

    image['filename'] = newname

def compute_filename(name):
    import string
    valid_chars = "-_.%s%s" % (string.ascii_letters, string.digits)
    return ''.join(c for c in name if c in valid_chars)

class VariableWidget(QWidget):
    def __init__(self, wtype, data=[], presets={}):
        QWidget.__init__(self)
        
        self.wtype = wtype
        self.presets = presets
        
        mainLayout = QVBoxLayout()
        self.setLayout(mainLayout)
        # self.bufferLayout.addStretch(0)
        mainLayout.setAlignment(Qt.AlignTop)
        
        self.components = []
        for d in data:
            c = self.add()
            c.load(d)
    
    def add(self):
        c = self.wtype("component" + str(len(self.components)+1), self.presets)
        self.layout().addWidget(c)
        c.remove.clicked.connect(lambda: self.remove(c))
        self.components.append(c)
        
        return c

    def remove(self, c):
        self.components.remove(c)
        for i, nqu in enumerate(self.components):
            nqu.setTitle("component" + str(i+1))
        self.layout().removeWidget(c)
        import sip
        sip.delete(c)
        # self.adjustSize()
    
    def value(self):
        return [c.value() for c in self.components]

class RadioWidget(QWidget):
    def __init__(self, parent):
        QWidget.__init__(self)
        mainLayout = QVBoxLayout()
        self.setLayout(mainLayout)
        self.buttongroup = QButtonGroup(self)
        self.loaded = None
        
    def setbuttons(self, buttons=[]):
        for button in buttons:
            self.buttongroup.addButton(button)
            self.layout().addWidget(button)
            if self.loaded:
                if button.text() == self.loaded:
                    button.setChecked(True)

    def load(self, checked=''):
        if checked:
            self.loaded = checked
    
    def value(self):
        if self.buttongroup.checkedButton():
            return self.buttongroup.checkedButton().text()
        else:
            return None
