#!/usr/bin/env python3
#  Author: Jonathan Michalon <michalon@igbmc.fr>, 2016-2017
#  Institut de Génétique et de Biologie Moléculaire et Cellulaire
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#  
 
import sys
from PyQt5.QtGui import *
from PyQt5.QtCore import *
from PyQt5.QtWidgets import *
import tarfile


class ItemSortWidget(QTableWidgetItem):
    
    def __lt__(self, other):
        w1 = self.tableWidget().cellWidget(self.row(), self.column())
        w2 = other.tableWidget().cellWidget(other.row(), other.column())
        return w1.layout().itemAt(0).widget().text() < w2.layout().itemAt(0).widget().text()

 
class GridFileApplication(QApplication):
 
    def __init__(self, args):
        QApplication.__init__(self, args)

        self.data = []
        self.samplepresets = {}
        self.bufferpresets = {}
        self.micropresets = {}
        self.infospresets = {}

        self.location = "./"
        self.getloc(args)
        self.tarfilename = "gridfiles.tar"
        self.imagesdir = "images/"

        # print(QStyleFactory.keys())
        # self.setStyle("GTK+")
        self.win = QMainWindow()
        self.win.setWindowTitle("Grid Files Manager")
        self.addWidgets()
        self.win.resize(1100, 500)
        self.win.show()
        self.slotLoad(0)
        
        QShortcut(QKeySequence("Ctrl+Q"), self.win, self.win.close)
 
    def addWidgets(self):
        centralWidget = QWidget()
        centralLayout = QVBoxLayout()
        centralWidget.setLayout(centralLayout)
        self.win.setCentralWidget(centralWidget)
        
        loc = QGroupBox("Locations")
        buttonLayout = QHBoxLayout()
        loc.setLayout(buttonLayout)
        self.locbutton = QPushButton()
        self.locbutton.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Preferred)
        self.locbutton.clicked.connect(self.slotLocation)
        buttonLayout.addWidget(self.locbutton)
        self.gridlabel = QLabel("Grid Files: " + self.location + self.tarfilename)
        self.gridlabel.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Preferred)
        self.gridlabel.setTextInteractionFlags(Qt.TextSelectableByMouse)
        buttonLayout.addWidget(self.gridlabel)
        self.imagelabel = QLabel()
        self.imagelabel.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Preferred)
        self.imagelabel.setTextInteractionFlags(Qt.TextSelectableByMouse)
        buttonLayout.addWidget(self.imagelabel)
        centralLayout.addWidget(loc)
        self.updateloc()
        data = QGroupBox("Data")
        buttonLayout = QHBoxLayout()
        loadbutton = QPushButton("Load",None)
        loadbutton.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Preferred)
        loadbutton.clicked.connect(self.slotLoad)
        buttonLayout.addWidget(loadbutton)
        savebutton = QPushButton("Save",None)
        savebutton.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Preferred)
        savebutton.clicked.connect(self.slotSave)
        buttonLayout.addWidget(savebutton)
        packbutton = QPushButton("Unpack/Repack",None)
        packbutton.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Preferred)
        packbutton.clicked.connect(self.slotPack)
        buttonLayout.addWidget(packbutton)
        data.setLayout(buttonLayout)
        centralLayout.addWidget(data)
        manage = QGroupBox("Manage")
        buttonLayout = QHBoxLayout()
        newbutton = QPushButton("New Grid",None)
        newbutton.clicked.connect(self.slotAdd)
        buttonLayout.addWidget(newbutton)
        newbutton = QLineEdit(None)
        newbutton.setPlaceholderText("Search, implementing !(not), &(and), |(or)")
        newbutton.textEdited.connect(self.slotSearch)
        newbutton.focusInEvent = lambda e: QTimer.singleShot(0, newbutton.selectAll)
        buttonLayout.addWidget(newbutton)
        backupbutton = QPushButton("Make backup",None)
        backupbutton.clicked.connect(self.slotBackup)
        buttonLayout.addWidget(backupbutton)
        manage.setLayout(buttonLayout)
        centralLayout.addWidget(manage)
        
        self.table = QTableWidget(0, 10)
        centralLayout.addWidget(self.table)
        self.table.horizontalHeader().setStretchLastSection(True)
        self.table.cellChanged.connect(self.slotItemChanged)
        self.table.setSortingEnabled(True)
        self.table.setHorizontalHeaderLabels(["Grid", "Sample", "Purif.", "Buffer", "Cryo. Cond.", "Microscope", "Atlas", "Image", "Infos", "Export"])
        self.table.resizeColumnsToContents()
        self.table.setSelectionMode(QAbstractItemView.NoSelection)
 
    def getconffile(self):
        import os
        return os.path.join(QDir.homePath(), '.gridfilesrc')
    
    def getloc(self, args):
        if len(args) == 2:
            self.location = args[1]
        else:
            import os
            conffile = self.getconffile()
            if os.access(conffile, os.R_OK):
                with open(conffile) as conf:
                    self.location = conf.readline()
            else:
                print ("No config file {}.".format(conffile))
    
    def updateloc(self):
        if not self.location.endswith('/'):
            self.location += '/'
        self.locbutton.setText("Base: " + self.location)
        self.gridlabel.setText("Grid Files: " + self.location + self.tarfilename)
        self.imagelabel.setText("Images: " + self.location + self.imagesdir)
        with open(self.getconffile(), 'w') as conf:
            conf.write(self.location)
    
    def slotLocation(self, dummy):
        (text, ok) = QInputDialog.getText(self.table, 'New location',
                     'What location do you want to use for data?\nSupports local directory or http (webdav) locations.',
                     QLineEdit.Normal, self.location)
        if ok and text:
            self.location = text
            self.updateloc()
            self.slotLoad(0)
    
    def toyamlandtar(self, data, name):
        import io
        import yaml
        
        stream = io.StringIO()
        yamldata = data.copy()
        if 'filename' in yamldata:
            del yamldata['filename']
        yaml.dump(yamldata, stream, default_flow_style=False)
        
        binary = io.BytesIO(stream.getvalue().encode('UTF-8'))
        ti = tarfile.TarInfo(name)
        ti.size = len(binary.getbuffer())
        return ti, binary
        
    def slotSave(self, ul=True):
        # ul: saving without uploading is usefull for unpack/repack
        import io
        import os
        
        if self.location.startswith('http://'):
            path = os.path.join(QDir.tempPath(), self.tarfilename)
        else:
            path = os.path.join(self.location, self.tarfilename)
            if not os.access(self.location, os.W_OK):
                os.makedirs(self.location)
        
        try:
            with tarfile.open(path, 'a') as tar:
                for gf in self.data:
                    ti, binary = self.toyamlandtar(gf, gf['filename'])
                    tar.addfile(ti, binary)
                ti, binary = self.toyamlandtar(self.bufferpresets, './presets/buffer.yaml')
                tar.addfile(ti, binary)
                ti, binary = self.toyamlandtar(self.micropresets, './presets/microscope.yaml')
                tar.addfile(ti, binary)
                ti, binary = self.toyamlandtar(self.samplepresets, './presets/sample.yaml')
                tar.addfile(ti, binary)
                ti, binary = self.toyamlandtar(self.infospresets, './presets/infos.yaml')
                tar.addfile(ti, binary)
        except Exception as e:
            from helpers import alert
            alert('Could not save data: ' + str(e))
        
        if self.location.startswith('http://'):
            from dav import push
            try:
                push(path, self.location)
            except Exception as e:
                from helpers import alert
                alert('Could not put data on server: ' + str(e))
                return
        
        return path # if caller wants to reuse (ie. backup)
        
    def addSetButton(self, slot, pos, name, data, withdate=False):
        if withdate:
            item = ItemSortWidget("");
            self.table.setItem(pos[0], pos[1], item)
            widget = QWidget()
            layout = QHBoxLayout()
            layout.setContentsMargins(0,0,0,0)
            widget.setLayout(layout)
            date = QLabel(data.get(name, {}).get('date', 'No date'))
            # init color of acq state here (!)
            if name == 'microscope' and 'microscope' in data and 'state' in data['microscope']:
                import microscope
                for button in microscope.MicroscopeDialog.getbuttons():
                    if button.text() == data[name]['state']:
                        date.setStyleSheet(button.styleSheet())
            layout.addWidget(date)
        button = QPushButton("Set",None)
        button.clicked.connect(lambda: slot(data))
        if withdate:
            layout.addWidget(button)
            self.table.setCellWidget(pos[0], pos[1], widget)
        else:
            self.table.setCellWidget(pos[0], pos[1], button)
        self.adjustColor(data, name, button)

    def addgrid(self, grid):
        self.data.append(grid)
        
        pos = len(self.data)-1
        self.table.setRowCount(pos+1)
        self.table.setSortingEnabled(False)
        
        # ask for a grid name if grid does not already exist
        if not grid:
            (text, ok) = QInputDialog.getText(self.table, 'Initial Grid Name', 'Which Grid Name do you want?')
            if ok and text:
                grid['name'] = text
        # compute a filename if still not done (migration or new grid)
        if 'filename' not in grid:
            from helpers import compute_filename
            grid['filename'] = compute_filename(grid['name'])
            grid['filename'] = './gridfiles/' + grid['filename'] + '.yaml'

        item = QTableWidgetItem(grid.get('name', ''));
        item.setData(Qt.UserRole, QVariant(pos))
        self.table.setItem(pos, 0, item)
        self.addSetButton(self.slotSample, (pos, 1), 'sample', grid)
        self.addSetButton(self.slotPurif,  (pos, 2), 'purif', grid)
        self.addSetButton(self.slotBuffer, (pos, 3), 'buffer', grid)
        self.addSetButton(self.slotCryo,   (pos, 4), 'cryo', grid, withdate=True)
        self.addSetButton(self.slotMicro,  (pos, 5), 'microscope', grid, withdate=True)
        self.addSetButton(self.slotAtlas,  (pos, 6), 'atlas', grid)
        self.addSetButton(self.slotImage,  (pos, 7), 'image', grid)
        self.addSetButton(self.slotInfos,(pos, 8), 'infos', grid)
        
        try:
            import pdf
            button = QPushButton("PDF",None)
            button.clicked.connect(lambda: self.slotPDF(grid))
        except Exception as e:
            print ("Cannot use export: {}".format(str(e)))
            button = QLabel("Not available",None)
        self.table.setCellWidget(pos, 9, button)
        
        self.table.setSortingEnabled(True)
        self.table.resizeColumnsToContents()
        
        
    def slotLoad(self, checked, dl=True):
        # checked is not used here but supplied by Qt, need to reference it for dl
        # dl parameter makes reloading after unpack/repack possible
        import yaml
        import os
        
        if self.location.startswith('http://'):
            path = os.path.join(QDir.tempPath(), self.tarfilename)
            if dl:
                from dav import pull
                try:
                    pull(path, self.location)
                except Exception as e:
                    from helpers import alert
                    alert('Could not get data from server: ' + str(e))
                    # remove broken file to avoid issues with tar in append mode
                    os.remove(path)
                    return
                
        else:
            path = os.path.join(self.location, self.tarfilename)
            # silently ignore nonexisting data
            if not os.access(path, os.R_OK):
                return

        self.data = []
        self.table.clearContents()
        try:
            with tarfile.open(path, 'r') as tar:
                for name in (name for name in sorted(set(tar.getnames())) if name.startswith('./gridfiles/') and name.endswith('yaml')):
                    try:
                        stream = tar.extractfile(name)
                    except KeyError:
                        pass
                    grid = yaml.load(stream)
                    grid['filename'] = name
                    # migration 'comment' => 'infos'
                    if grid.get('comment', {}):
                        grid['infos'] = grid.pop('comment')
                    self.addgrid(grid)
                for p in [('./presets/buffer.yaml', 'bufferpresets'),
                          ('./presets/microscope.yaml', 'micropresets'),
                          ('./presets/sample.yaml', 'samplepresets'),
                          ('./presets/infos.yaml', 'infospresets'),
                         ]:
                    try:
                        stream = tar.extractfile(p[0])
                        setattr(self, p[1], yaml.load(stream))
                    except KeyError:
                        pass
        except Exception as e:
            from helpers import alert
            alert('Could not load data: ' + str(e))
            
    
    def slotSample(self, grid):
        import sample
        d = sample.SampleDialog(self.win, grid.get('sample', {}), self.samplepresets)
        d.exec()
        if d.doupdate:
            grid.update ({'sample': d.data})
            self.samplepresets = d.presets
            self.adjustColor(grid, 'sample', self.win.sender())

    def slotPurif(self, grid):
        import purif
        d = purif.PurifDialog(self.win, grid.get('purif', {}))
        d.exec()
        if d.doupdate:
            grid.update ({'purif': d.data})
            self.adjustColor(grid, 'purif', self.win.sender())

    def slotBuffer(self, grid):
        import buffer
        d = buffer.BufferDialog(self.win, grid.get('buffer', {}), self.bufferpresets)
        d.exec()
        if d.doupdate:
            grid.update ({'buffer': d.data})
            self.bufferpresets = d.presets
            self.adjustColor(grid, 'buffer', self.win.sender())
            
    def slotCryo(self, grid):
        import cryo
        d = cryo.CryoDialog(self.win, grid.get('cryo', {}))
        d.exec()
        if d.doupdate:
            grid.update ({'cryo': d.data})
            self.adjustColor(grid, 'cryo', self.win.sender())
            self.win.sender().parent().layout().itemAt(0).widget().setText(d.data['date'])
            self.table.resizeColumnsToContents()

    def slotMicro(self, grid):
        import microscope
        d = microscope.MicroscopeDialog(self.win, grid.get('microscope', {}), self.micropresets)
        d.exec()
        if d.doupdate:
            grid.update ({'microscope': d.data})
            self.micropresets = d.presets
            self.adjustColor(grid, 'microscope', self.win.sender())
            date = self.win.sender().parent().layout().itemAt(0).widget()
            date.setText(d.data['date'])
            if d.data['state']:
                date.setStyleSheet(d.radios.buttongroup.checkedButton().styleSheet())
            self.table.resizeColumnsToContents()

    def slotAtlas(self, grid):
        import atlas
        d = atlas.AtlasDialog(self.win, self.location, self.imagesdir, grid['filename'], grid.get('atlas', {}))
        d.exec()
        if d.doupdate:
            grid.update ({'atlas': d.data})
            self.adjustColor(grid, 'atlas', self.win.sender())

    def slotImage(self, grid):
        import image
        d = image.ImageDialog(self.win, self.location, self.imagesdir, grid['filename'], grid.get('image', {}))
        d.exec()
        if d.doupdate:
            grid.update ({'image': d.data})
            self.adjustColor(grid, 'image', self.win.sender())

    def slotInfos(self, grid):
        import infos
        d = infos.InfosDialog(self.win, grid.get('infos', {}), self.infospresets)
        d.exec()
        if d.doupdate:
            grid.update ({'infos': d.data})
            self.infospresets = d.presets
            self.adjustColor(grid, 'infos', self.win.sender())

    def slotPDF(self, grid):
        from pdf import genpdf
        from helpers import compute_filename
        defaultname = compute_filename(grid.get('name', '')) + ".pdf"
        (filename, ext) = QFileDialog.getSaveFileName(self.table, "Select PDF file", defaultname, "PDF document (*.pdf);;All Files (*)")
        if filename:
            genpdf(grid, filename, self.location, self.imagesdir)

    def slotPack(self):
        import os
        self.slotSave(ul=False)
        import dataonfs
        loc = QDir.tempPath() if self.location.startswith('http://') else self.location
        d = dataonfs.DataOnFS(self.win, os.path.join(loc, self.tarfilename))
        d.exec()
        self.slotLoad(0, dl=False)
    
    def slotAdd(self):
        self.addgrid({})

    def rename_grid(self, grid, newname):
        from helpers import alert, moveimage
        if not newname:
            alert("Name empty!")
            return
        for gf in self.data:
            if gf['name'] == newname:
                alert("Grid '%s' already exists" % newname)
                return False
        for image in ['atlas', 'image']:
            if image in grid and 'filename' in grid[image]:
                moveimage(self.location, self.imagesdir, grid[image], newname)
        self.slotSave()
        alert("Rename to '%s' applied, data saved." % newname)
        return True
    
    def slotItemChanged(self, x, y):
        if y == 0: # name column
            pos = self.table.item(x, y).data(Qt.UserRole)
            newname = self.table.item(x, y).text()
            if newname != self.data[pos]['name']:
                if self.rename_grid(self.data[pos], newname):
                    self.data[pos]['name'] = self.table.item(x, y).text()
                else:
                    self.table.item(x, y).setText(self.data[pos]['name'])
            
    def slotSearch(self, text):

        # search a match in the whole grid data
        def dolinesearch(text, data):
            # keep elements to search in
            todo = [data]
            # while there is something to search in…
            while todo:
                # pick one
                val = todo.pop()
                # switch on type, add to todo or search directly 
                if type(val) is dict:
                    todo += [v for v in iter(val.values())]
                elif type(val) is list:
                    todo += [v for v in val]
                elif type(val) is int or type(val) is float:
                    v = text
                    try:
                        if type(val) is int:
                            v = int(text)
                        if type(val) is float:
                            v = float(text)
                    except ValueError:
                        pass
                    if v == val:
                        return True
                elif type(val) is str:
                    if text in val:
                        return True
            return False

        # recursive expression search
        def doexprsearch(text, data):
            begin, sep, end = text.partition('&')
            if sep:
                return doexprsearch(begin, data) and doexprsearch(end, data)

            begin, sep, end = text.partition('|')
            if sep:
                return doexprsearch(begin, data) or doexprsearch(end, data)

            if text and text[0] == '!':
                return not doexprsearch(text[1:], data)

            return dolinesearch(text, data)

        # loop on grids
        for i, d in enumerate(self.data):
            match = doexprsearch(text, d)

            # (de)activate the various elements of the current grid according
            # to the presence of the search pattern
            for j in range(self.table.rowCount()):
                if self.table.item(j, 0).data(Qt.UserRole) == i:
                    for k in range(self.table.columnCount()):
                        item = self.table.item(j, k)
                        if item:
                            if match:
                                item.setFlags(item.flags() |  Qt.ItemIsEnabled)
                            else:
                                item.setFlags(item.flags() & ~Qt.ItemIsEnabled)
                        widget = self.table.cellWidget(j, k)
                        if widget:
                            widget.setEnabled(match)
    
    def slotBackup(self, dummy):
        from helpers import alert, getimage
        import os
        import shutil
        directory = QFileDialog.getExistingDirectory(self.table, "Select output directory")
        if directory:
            now = QDateTime.currentDateTime()
            backup = os.path.join(directory, "gridfiles-backup-" + now.toString("yyyyMMddhhmmss"))
            alert("Will save and copy all data to {}.".format(backup))
            tarfile = self.slotSave()
            QDir().mkpath(backup)
            QFile.copy(tarfile, os.path.join(backup, self.tarfilename))
            # copy all referenced images
            images = []
            for grid in self.data:
                if 'atlas' in grid and 'filename' in grid['atlas']:
                    images.append(grid['atlas']['filename'])
                if 'image' in grid and 'filename' in grid['image']:
                    images.append(grid['image']['filename'])
            QDir().mkpath(os.path.join(backup, self.imagesdir))
            for image in images:
                out_file = open(os.path.join(backup, self.imagesdir, image), 'wb')
                data = getimage(self.location, self.imagesdir, image)
                shutil.copyfileobj(data, out_file, length=16*1024*1024)
        
    def adjustColor(self, data, key, obj):
        if key in data and data[key]:
            if 'valid' in data[key] and data[key]['valid']:
                obj.setStyleSheet("background-color: green")
            else:
                obj.setStyleSheet("")
        else:
            obj.setStyleSheet("background-color: red")

            
if __name__ == "__main__":
    app = GridFileApplication(sys.argv)
    sys.exit(app.exec_())
