#  Author: Jonathan Michalon <michalon@igbmc.fr>, 2016-2017
#  Institut de Génétique et de Biologie Moléculaire et Cellulaire
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#  
from reportlab.platypus import (
    ListFlowable,
    Frame, 
    Paragraph
)
from reportlab.lib.styles import getSampleStyleSheet
from reportlab.pdfgen import canvas as pdfcanvas
from reportlab.lib.pagesizes import A4

styles = getSampleStyleSheet()
normal = styles["Normal"]
normal.fontName = "Courier"
heading = styles['Heading1']
import copy
heading2 = copy.copy(heading)
heading2.fontSize = 12
heading2.leading = 10
heading2.spaceBefore = 7

canvas = None
width, height = A4
edgel = width/15
edger = width*14/15

# helper to add a field as Paragraph
def appendifexists(l, data, klass, field, br=False):
    fields = klass.fields
    if field in data and field in fields:
        l.append(Paragraph('{}:{} {} {}'.format(fields[field][1], '<br />' if br else '', data[field], fields[field][2]), normal))

def drawsample(data):
    from sample import SampleDialog as dialog
    elems = []
    elems.append(Paragraph('Sample:', heading))
    comp = []
    if 'components' in data:
        elems.append(Paragraph('Components:', normal))
        for c in data['components']:
            key, value = list(c.items())[0]
            comp.append(Paragraph("{}: {}".format(value[0], value[1]), normal))
    elems.append(ListFlowable(comp, bulletType='bullet', start=' -'))
    appendifexists(elems, data, dialog, 'concentration')
    f = Frame(edgel, height-290, 250, 170, showBoundary=0)
    f.addFromList(elems, canvas)

def drawpurif(data):
    from purif import PurifDialog as dialog
    elems = []
    elems.append(Paragraph('Purification:', heading))
    appendifexists(elems, data, dialog, 'descr', br=True)
    f = Frame(edgel, 280, 140, 230, showBoundary=0)
    f.addFromList(elems, canvas)
    
def drawbuffer(data):
    from buffer import BufferDialog as dialog
    elems = []
    elems.append(Paragraph('Buffer:', heading))
    appendifexists(elems, data, dialog, 'ph')
    elems.append(Paragraph('Components:', normal))
    comp = []
    if 'components' in data:
        for c in data['components']:
            key, value = list(c.items())[0]
            comp.append(Paragraph("{}: {} {}".format(value[0], value[1], value[2]), normal))
    elems.append(ListFlowable(comp, bulletType='bullet', start=' -'))
    f = Frame(width/2-70, 280, 140, 230, showBoundary=0)
    f.addFromList(elems, canvas)

def drawcryo(data):
    from cryo import CryoDialog as dialog
    elems = []
    elems.append(Paragraph('Cryo. Cond.:', heading))
    appendifexists(elems, data, dialog, 'date')
    elems.append(Paragraph('Glow Discharge:', heading2))
    appendifexists(elems, data, dialog, 'charging')
    appendifexists(elems, data, dialog, 'duration')
    appendifexists(elems, data, dialog, 'pressure')
    appendifexists(elems, data, dialog, 'tafter')
    elems.append(Paragraph('Vitrification:', heading2))
    appendifexists(elems, data, dialog, 'temperature')
    appendifexists(elems, data, dialog, 'humidity')
    appendifexists(elems, data, dialog, 'time')
    appendifexists(elems, data, dialog, 'force')
    appendifexists(elems, data, dialog, 'quantity')
    
    f = Frame(edger-140, 280, 140, 230, showBoundary=0)
    f.addFromList(elems, canvas)

def drawmicro(data):
    from microscope import MicroscopeDialog as dialog
    elems = []
    elems.append(Paragraph('Microscope:', heading))
    appendifexists(elems, data, dialog, 'date')
    appendifexists(elems, data, dialog, 'grid')
    appendifexists(elems, data, dialog, 'microscope')
    appendifexists(elems, data, dialog, 'voltage')
    appendifexists(elems, data, dialog, 'camera')
    appendifexists(elems, data, dialog, 'magnification')
    appendifexists(elems, data, dialog, 'mindefocus')
    appendifexists(elems, data, dialog, 'maxdefocus')
    appendifexists(elems, data, dialog, 'state')
    
    f = Frame(edger-220, height-290, 220, 170, showBoundary=0)
    f.addFromList(elems, canvas)

def drawimage(data, location, imagesdir, x, y):    
    if 'filename' in data:
        try:
            from PIL import Image
            from helpers import getimage
            data = getimage(location, imagesdir, data['filename'])
            image = Image.open(data)
            canvas.drawInlineImage(image, x, y, width=370, height=370)
        except Exception as e:
            print ("Cannot draw image in PDF: {}".format(str(e)))
    
def drawinfos(data):
    from infos import InfosDialog as dialog
    elems = []
    elems.append(Paragraph('Infos:', heading))
    appendifexists(elems, data, dialog, 'comment', br=True)
    elems.append(Paragraph('Users:', normal))
    comp = []
    if 'users' in data:
        for c in data['users']:
            key, value = list(c.items())[0]
            comp.append(Paragraph("{} / {}".format(value[0], value[1]), normal))
    elems.append(ListFlowable(comp, bulletType='bullet', start=' -'))
    f = Frame(edgel, 130, width*13/15, 120, showBoundary=0)
    f.addFromList(elems, canvas)

def genpdf(grid, filename, location, imagesdir):
    
    global canvas
    
    canvas = pdfcanvas.Canvas(filename, pagesize=A4)
    
    canvas.setFont('Times-Bold', 36)
    canvas.drawCentredString(width/2.0, height-70, grid['name'])
    canvas.setFont('Times-Roman', 9)

    canvas.drawString(edgel,height-40, "Grid File")
    
    drawsample(grid.get('sample', {}))
    drawpurif(grid.get('purif', {}))
    drawbuffer(grid.get('buffer', {}))
    drawcryo(grid.get('cryo', {}))
    drawmicro(grid.get('microscope', {}))
    drawinfos(grid.get('infos', {}))
    
    canvas.showPage()

    if 'atlas' in grid or 'image' in grid:
        canvas.setFont('Times-Bold', 36)
        canvas.drawCentredString(width-(width-width/15-370)/2, height-70, grid['name'])
        canvas.setFont('Times-Roman', 9)

        if 'atlas' in grid and 'filename' in grid['atlas']:
            canvas.setFont('Times-Bold', 24)
            canvas.drawString(edger-130, height-135, 'Atlas')
            canvas.setFont('Times-Roman', 9)
            drawimage(grid.get('atlas', {}), location, imagesdir, edgel, height-40-370)
        if 'image' in grid and 'filename' in grid['image']:
            canvas.setFont('Times-Bold', 24)
            canvas.drawString(edgel+20, height/2-50-24, 'Example')
            canvas.drawString(edgel+20, height/2-75-24, 'Image')
            canvas.setFont('Times-Roman', 9)
            drawimage(grid.get('image', {}), location, imagesdir, edger-370, 40)
        
        canvas.showPage()
    canvas.save()
